#include "sockutil.h"
#include <wsnwlink.h>

#define PRINT_ERROR_AND_RETURN(s,err) {\
    printf("Error: %d in %s\n", err, s);\
    return(err);\
}

//
// Initialize WinSock library
//
int WinSockInitialize()
{
    int      err;
    WORD     wVersionRequired;
    WSADATA  wsaData;
    wVersionRequired = MAKEWORD(1,1);
    err = WSAStartup(wVersionRequired, &wsaData);
    if (err != 0)
        exit(1);
    return(0);
}

//
// Get local or remote host name
//
u_long GetHostAddress(LPSTR pszHostName)
{
    struct hostent FAR *pHostAddr;
#ifdef _DEBUG
    struct in_addr in;
#endif

    //
    // get local host address
    //
    pHostAddr = gethostbyname ( pszHostName );
    if (pHostAddr == NULL)
    {
        printf("gethostbyname error: %ld for %s\n",
                WSAGetLastError(), pszHostName);
        return(0);
    }
#ifdef _DEBUG
    else // print out details
    {
        printf("Host Name: %s\n",       pHostAddr->h_name);
        printf("Address Type: %x\n",    pHostAddr->h_addrtype);
        printf("Address Length: %d\n",  pHostAddr->h_length);
        printf("Host Address: %lx \n",  pHostAddr->h_addr);
        in.s_addr = *((u_long *)pHostAddr->h_addr);
        printf("Dotted Address: %s\n", inet_ntoa(in));
    }
#endif
    return *((u_long *)pHostAddr->h_addr);
}

//
// Get Local Host Address
// 
// This may causes lookup in %SYSTEMROOT%\drivers\etc\hosts file 
u_long GetLocalHostAddress()
{
    char   szHostName [ MAX_PATH ];
    u_long ulHostAddr;
    if (gethostname (szHostName, sizeof(szHostName)) != SOCKET_ERROR)
    {
#ifdef _DEBUG
        printf("Local Host Name: %s\n", szHostName);
#endif
        ulHostAddr = GetHostAddress ( szHostName );
    }
    else
    {
#ifdef _DEBUG
        printf("gethostname error: %ld\n", WSAGetLastError());
#endif
        return(0);
    }
    return (ulHostAddr);
}

//
// Get port number for a local service
//
// This causes lookup in %SYSTEMROOT%\drivers\etc\services file 
u_short GetPortForService(LPSTR pszServiceName, LPSTR pszProtocol)
{
    u_short usPort;
    struct servent FAR* pServAddr;
    pServAddr = getservbyname(pszServiceName, pszProtocol);
    if (pServAddr != NULL)
    {
        // Print Service Details
#ifdef _DEBUG
        printf("\nService Name: %s\n",     pServAddr->s_name);
        printf("Service Port: %d\n",     ntohs(pServAddr->s_port));
        printf("Service Protocol: %s\n",   pServAddr->s_proto);
#endif
        usPort = pServAddr->s_port;
    }
    else
    {
        printf("getservbyname error: %ld\n", WSAGetLastError());
        return ((u_short)-1);
    }
    return(usPort);
}

//
// Create a socket
//
SOCKET CreateSocket (int type)
{
    SOCKET soc;

    // Create a stream socket AF_INET (TCP, UDP) address family
    soc = socket (AF_INET, type, 0);

    return (soc);
}

//
// Bind a socket to the host address and service port
//
int BindSocketToAddr (
        int    iBindType,
        SOCKET soc,
        LPSTR  pszServiceName,
        LPSTR  pszProtocolName,
        int    iPortNumber,
        BOOL   fClient)
{
    struct sockaddr_in LocalAddr;
    int                err;
    u_long             ulHostAddress;
    u_short            usPortNumber;

    //
    // For TCP/IP connections, Client asks the system to allocate
    // a port and binds to INADDR_ANY
    //
    if (fClient)
    {
        iBindType = DEFAULT_PORT;
    }
    memset ( &LocalAddr, 0x00, sizeof(LocalAddr));
    err = SOCKET_ERROR;

    switch (iBindType)
    {
    case DEFAULT_PORT:
        // Bind to local host address and get system to allocate a port number
        LocalAddr.sin_family = AF_INET;
        LocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);
        LocalAddr.sin_port = 0;
        break;
    case FIXED_PORT:
        // Bind to the local host address and a fixed port
        if ((ulHostAddress = GetLocalHostAddress()) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetLocalHost", err);
        }
        LocalAddr.sin_family = AF_INET;
        LocalAddr.sin_addr.s_addr = ulHostAddress ;
        LocalAddr.sin_port = htons((short) iPortNumber);
        break;
    case LOCAL_ADDR:
        // Bind to the local host address and a port for local service
        if ((ulHostAddress = GetLocalHostAddress()) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetLocalHost error", err);
        }
        else
        {
            usPortNumber = GetPortForService(pszServiceName, pszProtocolName);
        }
        LocalAddr.sin_family = AF_INET;
        LocalAddr.sin_addr.s_addr = ulHostAddress ;
        LocalAddr.sin_port = usPortNumber;
        break;
    };

    //
    // Bind the socket to the local address
    //
    err = bind(soc, (struct sockaddr *)&LocalAddr, sizeof(LocalAddr));
    if (err==SOCKET_ERROR)
    {
        PRINT_ERROR_AND_RETURN("bind()", WSAGetLastError());
    }
#ifdef _DEBUG
    else
        printf("Successful Binding\n");
#endif

    return (err);
}

//
// Connect to a local or remote Server
//
int ConnectToServer  (
        int    iBindType,
        SOCKET soc,
        LPSTR  pszHostName,
        LPSTR  pszServiceName,
        LPSTR  pszProtocolName,
        int    iPortNumber)
{
    struct sockaddr_in RemoteAddr;
    int err;
    u_long ulHostAddress;
    u_short usPortNumber;

    memset (&RemoteAddr, 0x00, sizeof(RemoteAddr));
    err = SOCKET_ERROR;

    switch (iBindType)
    {
    case DEFAULT_PORT:
        printf("Error: Invalid option in ConnectToServer\n");
        return(SOCKET_ERROR);
        break;
    case FIXED_PORT:
        // Connect to local Server with a fixed port
        if ((ulHostAddress = GetLocalHostAddress()) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetHostAddress", err);
        }
        RemoteAddr.sin_family = AF_INET;
        RemoteAddr.sin_addr.s_addr = ulHostAddress ;
        RemoteAddr.sin_port = htons((short) iPortNumber);
        break;
    case LOCAL_ADDR:
        // Connect to a local Server. Get it's port at run-time
        if ((ulHostAddress = GetLocalHostAddress()) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetLocalHost error", err);
        }
        else
        {
            usPortNumber = GetPortForService(pszServiceName, pszProtocolName);
        }
        RemoteAddr.sin_family = AF_INET;
        RemoteAddr.sin_addr.s_addr = ulHostAddress ;
        RemoteAddr.sin_port = usPortNumber;
        break;
    case REMOTE_ADDR:
        // Connect to a remote Server. Get it's port at run-time.
        if ((ulHostAddress = GetHostAddress(pszHostName)) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetLocalHost error", err);
        }
        else
        {
            usPortNumber = GetPortForService(pszServiceName, pszProtocolName);
        }

        RemoteAddr.sin_family = AF_INET;
        RemoteAddr.sin_addr.s_addr = ulHostAddress;
        RemoteAddr.sin_port = usPortNumber;
        break;
    case REMOTE_PORT:
        // Connect to a remote Server, use the provided port-number
        if ((ulHostAddress = GetHostAddress(pszHostName)) == 0)
        {
            PRINT_ERROR_AND_RETURN("GetLocalHost error", err);
        }

        RemoteAddr.sin_family = AF_INET;
        RemoteAddr.sin_addr.s_addr = ulHostAddress;
        RemoteAddr.sin_port = htons((short) iPortNumber);
        break;
    };

    //
    // Connect the Client socket with the remote Server
    //
    err = connect(soc, (struct sockaddr *)&RemoteAddr, sizeof(RemoteAddr));
    if (err==SOCKET_ERROR)
    {
        PRINT_ERROR_AND_RETURN("connect()", WSAGetLastError());
    }
#ifdef _DEBUG
    else
        printf("Successful connection\n");
#endif

    return (err);
}

// Macros for printing hex numbers

#define ctoh(ch) isdigit(ch)?(ch - '0'):(10 + toupper(ch) - 'A')

#define PrintHex(ptr, size) {\
    for (i=0; i < size; i++)\
        printf("%x", ptr[i]);\
    printf("\n");\
}

void atoh(char *dest, char *src, int size)
{
    int i, j;
    for (i = 0, j = 0; i < size ; i++, j += 2)
    {
        dest[i]  = ctoh(src[j  ]);   // upper nibble
        dest[i] <<= 4;
        dest[i] |= ctoh(src[j+1]);   // lower nibble
    }
}

