//
// Windows Socket Client to fetch files from websites
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sockutil.h"
#include "parsecmd.h"

#define LF 10
#define CR 13

#define BUFFER_SIZE 16384

//
// wsrv - the main routine
//
BOOL wsrv (char * szRootDir, int iPortNumber, BOOL bVerbose)
{
    // URL data
    char szProtocol [MAX_PATH] = "tcp";
    char szServName [MAX_PATH] = "http";
    char szHostName [MAX_PATH] = "localhost";
    char szPathName [MAX_PATH] = "/";

    char ClientAddr [MAX_PATH];
    int iAddrLen;

    char szRequest [BUFFER_SIZE];
    char szBuffer [BUFFER_SIZE];
    char szHeader [BUFFER_SIZE];
    int iHttpProtocol = 0;

    SOCKET socketHandle, socketHandleAccept;
    int socketResult;

    int fileHandle;
    int fileResult;
    struct _stat fileStatus;
    
    char szFileType [MAX_PATH] = "text/html";
    char szFileSize [MAX_PATH] = "0";
    char szFileTime [MAX_PATH] = "Thu, 01 Jan 1970 00:00:00";
    time_t tDateTime;
    char szDateTime [MAX_PATH] = "Thu, 01 Jan 1970 00:00:00";

    BOOL bLoopServer = TRUE;
    int i;

    // initialize WinSock library
    WinSockInitialize();

    // Create a stream socket
    socketHandle = CreateSocket (SOCK_STREAM);
    if (socketHandle == INVALID_SOCKET)
    {
        WSACleanup();
        return(1);
    }

    if ((iPortNumber == 0) || (iPortNumber == 80))
        socketResult = BindSocketToAddr (LOCAL_ADDR, socketHandle, szServName, szProtocol, 0, FALSE);
    else
        socketResult = BindSocketToAddr (FIXED_PORT, socketHandle, NULL, szProtocol, iPortNumber, FALSE);
    if (socketResult)
    {
        closesocket(socketHandle);
        WSACleanup();
        return(1);
    }

    // create an incoming client connect request queue
    socketResult = listen (socketHandle, CLIENT_QUEUE_SIZE);
    if (socketResult)
    {
        printf("listen error: %d\n", WSAGetLastError());
        closesocket(socketHandle);
        WSACleanup();
        return(1);
    }

    // loop forever
    while (bLoopServer)
    {
        // wait for clients to call connect
        iAddrLen = sizeof(ClientAddr);
        memset (&ClientAddr, 0x00, iAddrLen);
        socketHandleAccept = accept (socketHandle, (struct sockaddr FAR *) &ClientAddr, &iAddrLen);
        if (socketHandleAccept != INVALID_SOCKET)
        {
            // read clients request
            socketResult = recv (socketHandleAccept, szRequest, sizeof (szRequest), 0);
            if (socketResult == SOCKET_ERROR)
            {
                printf ("Error: %d in recv\n", WSAGetLastError());
                closesocket (socketHandleAccept);
            }
            szRequest [socketResult - 2] = '\0';
            if (bVerbose)
                printf ("%s\n", szRequest);

            // Write back a reply
            if ((strncmp (szRequest, "GET", 3) == 0) ||
                (strncmp (szRequest, "HEAD", 4) == 0))
            {
                // compose pathname from root and client request
                strcpy (szPathName, szRootDir);
                strcat (szPathName, (szRequest + 5));
                for (i = 0; i < (int) strlen(szPathName); i++)
                {
                    if (szPathName[i] == '/')
                        szPathName[i] = '\\';
                    if ((szPathName[i] == ' ') || (szPathName[i] == '\n'))
                    {
                        szPathName[i] = '\0';
                        break;
                    }
                }

                // when directory, search for index.html
                if (szPathName [strlen(szPathName) - 1] == '\\')
                    strcat (szPathName, "index.html");

                // store http protocol-level
                if (strstr (szRequest, "HTTP/1.0") != NULL)
                    iHttpProtocol = 10;
                else if (strstr (szRequest, "HTTP/1.1") != NULL)
                    iHttpProtocol = 11;
                else
                    iHttpProtocol = 0;

                // check existance of requested file and open it
                if ((fileHandle = _open (szPathName, _O_RDONLY | _O_BINARY)) !=  -1)
                {
                    // get data associated with file
                    fileResult = _fstat( fileHandle, &fileStatus);

                    if (fileResult == 0)
                    {
                        sprintf (szFileSize, "%ld", fileStatus.st_size);
                        sprintf (szFileTime, "%s", ctime (&fileStatus.st_mtime));
                        szFileTime[strlen(szFileTime) - 1] = '\0';
                    }
                    
                    // set Mime file-type
                    if (strstr (szPathName, ".html") != NULL)
                        strcpy (szFileType, "text/html");
                    else if (strstr (szPathName, ".htm") != NULL)
                        strcpy (szFileType, "text/html");
                    else if (strstr (szPathName, ".txt") != NULL)
                        strcpy (szFileType, "text/plain");
                    else if (strstr (szPathName, ".gif") != NULL)
                        strcpy (szFileType, "image/gif");
                    else if (strstr (szPathName, ".jpeg") != NULL)
                        strcpy (szFileType, "image/jpeg");
                    else if (strstr (szPathName, ".png") != NULL)
                        strcpy (szFileType, "image/png");
                    else
                        strcpy (szFileType, "text/plain");                    
                    
                    // if HTTP, send header
                    if (iHttpProtocol > 0)
                    {
                        strcpy (szBuffer, "");
                        strcat (szBuffer, "HTTP/1.0 200 OK\r\n");
                        strcat (szBuffer, "Server: wsrv 1.0\r\n");
                        time (&tDateTime);
                        sprintf (szDateTime, "%s", ctime (&tDateTime));
                        szDateTime[strlen(szDateTime) - 1] = '\0';
                        sprintf (szHeader, "Date: %s\r\n", szDateTime);
                        strcat (szBuffer, szHeader);
                        sprintf (szHeader, "Last-Modified: %s\r\n", szFileTime);
                        strcat (szBuffer, szHeader);
                        sprintf (szHeader, "Content-length: %s\r\n", szFileSize);
                        strcat (szBuffer, szHeader);
                        sprintf (szHeader, "Content-type: %s\r\n", szFileType);
                        strcat (szBuffer, szHeader);
                        strcat (szBuffer, "\r\n");

                        // send header
                        send (socketHandleAccept, szBuffer, strlen (szBuffer), 0);
                    }

                    // send file content only for GET, not for HEAD request
                    if (strncmp (szRequest, "GET", 3) == 0)
                    {
                        while ((fileResult = _read(fileHandle, szBuffer, sizeof (szBuffer))) > 0)
                        {
                            send (socketHandleAccept, szBuffer, fileResult, 0);
                        }
                    } // end if GET

                    // close file
                    _close (fileHandle);
                }
                else // if file exists
                {
                    // send not-found error message
                    strcpy (szBuffer, "");
                    strcat (szBuffer, "HTTP/1.0 404 File Not Found\r\n");
                    strcat (szBuffer, "Server: wsrv 1.0\r\n");
                    sprintf (szHeader, "Date: %s\r\n", szDateTime);
                    strcat (szBuffer, szHeader);
                    strcat (szBuffer, "Content-length: 111\r\n");
                    strcat (szBuffer, "Content-type: text/html\r\n");
                    strcat (szBuffer, "\r\n");

                    strcat (szBuffer, "<HTML><HEAD><TITLE>File Not Found</TITLE></HEAD>\r\n");
                    strcat (szBuffer, "<BODY><H1>HTTP Error 404: File Not Found</H1></BODY></HTML>\r\n");

                    // send header and body
                    send (socketHandleAccept, szBuffer, strlen (szBuffer), 0);
                }
            }
        }
        // all done, disconnect the client
        closesocket (socketHandleAccept);
    }

    // close the socket
    closesocket (socketHandle);

    // cleanup and return
    WSACleanup();

    return TRUE;
}

//
// MAIN - the program
//
int main (int argc, char **argv)
{
    // URL data
    BOOL bVerbose = FALSE;
    char szRootDir [MAX_PATH] = "";
    int iPortNumber = 80;

    // Parse command line
    ParseCmdLine (argc, argv, (char *) szRootDir, &iPortNumber, &bVerbose);

    if (!wsrv((char *) szRootDir, iPortNumber, bVerbose))
    {
        return 1;
    }

    return 0;
}

